<?
/* --------------------------------------------------------------------
   Copyright (c) 2008-2012 Daniel Pedemont (dan@pedemont.net)
   -------------------------------------------------------------------- */
/**
 * stream src_relay
 *
 * stream socket endpoint connection operator
 *
 * @link http://dev.pedemont.com/sonic/
 * @link http://dev.pedemont.com/sonic/downloads/src/lib/streams/sonic.stream.src_relay.php
 * @package sonic_stream
 * @version 0.2.14
 * @abstract connection forwarder
*/

define('SONIC_SR_VERSION', '0.2.15');

//NB, function names in rotate mode have no protection,
//    so dont use generic function names for clash safety

class SONIC_STREAM_src_relay extends SONIC_stream
{
/*
                       sock_in         src_relay        sock_out
                      |       |       |         |       |      |
                      |       |       |         |       |      |
                  --> |  udp/ | ----> |[src_any]| ----> | udp/ | -->
                      |  tcp  |       | mux_any |       | tcp  |
             ip:port  |       |       |         |       |      |  dst_host:dst_port
                      |       |       |         |       |      |
                  <-- |  tcp/ | <---- |[mux_tcp]| <---- | tcp  | <--
                      |  udp  |       |         |       |      |
   mux_host:mux_port  |       |       |         |       |      |

//

                  [ tcp-1 ] <<---->> [ src relay ] <<--->> [ tcp-2 ]

                  [ udp/  ]
                  [ tcp-1 ] >>--.                   
                                 `>> [           ]
                                     [ src relay ] <<--->> [ tcp-2 ]
                  [ udp/  ]      ,<< [           ]
                  [ tcp-3 ] <<--'

                  [ tcp/  ]
                  [ udp-1 ] >>---->> [ src relay ] >>--->> [ udp-2 ]

*/
  /**
   * destination endpoint resource socket var
   *
   * @var resource
  */
  var $sock_out = NULL;
  /**
   * mux'ed resource socket output var
   *
   * @var resource
  */
  var $sock_mux_out = NULL;

  /**
   * signal termination handler
   *
   * @param integer $sig posix process signal
  */
  function _sonic_stream_sr_handler($sig = NULL)
  {
    switch ($sig)
    { 
      case SIGTERM:
      case SIGQUIT:
      case SIGINT:
        if (($this->pid == 0 && $this->pid_child == 0) OR ($this->pid == $this->pid_child)) // rotate mode
        { 
          sonic_daemon_thread_shutdown($sig);
          $this->rotate_shutdown = true;
        }
        else
        { 
          $this->sonic_stream_sr_stop();
          $this->sonic_stream_buffer();
          exit();
        }
      break;

      default:
        die('src_relay says WTF');
    }
  }

  /**
   * main setup
   *
   * load signal handers and 
   * poll socket_accept listening socket for connected client then return
   *
   * @see SONIC_STREAM_src_relay::sonic_stream_main()
   * @see SONIC_STREAM_src_relay::sonic_stream_sr_prelog()
   * @param resource third party support (library mode) override, listening resource socket
  */
  function sonic_stream_sr_start($sock_in = null)
  {
    if ($this->pcntl_loaded)
    {
      pcntl_signal(SIGTERM, array(&$this, '_sonic_stream_sr_handler'));
      pcntl_signal(SIGQUIT, array(&$this, '_sonic_stream_sr_handler'));
      pcntl_signal(SIGINT, array(&$this, '_sonic_stream_sr_handler'));
    }

    if ($sock_in)                       //3rd party compat
      $this->sock = &$sock_in;
    else if ($this->sock)               //sonic daemon inherit preload
    { 
      $sock_in = &$this->sock;
    }
    else if ($this->config['socket'])   //despro search
    { 
      $sock_in = &$this->config['socket'];
    }
    else
      die('! '.$this->sonic_stream_remote_prelog().' main() failed: reason: no socket available'."\n");

//    var_dump($this->config);

    if (($_bs = 0 + $this->config['byte_size']) > 0) {}
    else { $_bs = 2048; }
    define('SONIC_RELAY_BYTE_SIZE', $_bs); unset($_bs);

    $stream_config = $this->config['stream_config'];

    if ($this->verbose > 0)
    {
      if ($msg = $stream_config['msg_listen'])
        $msg .= "\n";
      else
        $msg = "listening for connection until close\n";

      echo '= '.$this->sonic_stream_sr_prelog().' ';
      echo $msg;
    }

    if ($this->config['socket_type'] == SOCK_STREAM)
    {
      $this->sonic_stream_buffer();

      if ($_sock_in = $this->sonic_stream_socket_accept($sssa_msg))
      { 
        if ($this->verbose > 0)
          echo '= '.$this->sonic_stream_sr_prelog().' '.$sssa_msg."\n";
      }
      else
        die('! '.$this->sonic_stream_sr_prelog().' main() failed: reason: '.$sssa_msg."\n");

      //WARN, overwrite root with accept socket
      $this->sock = $_sock_in;
      unset($_sock_in);
      $sock_in = &$this->sock;
    }
  }

  /**
   * main shutdown/finish
   *
   * closes in and out resource socket variables and echos trace
   *
   * @see SONIC_STREAM_src_relay::sonic_stream_main()
   * @see SONIC_STREAM_src_relay::sonic_stream_sr_prelog()
  */
  function sonic_stream_sr_stop()
  {
    if ($this->verbose > 0)
    {
      $stream_config = $this->config['stream_config'];

      if ($msg_close = $stream_config['msg_close'])
        $msg_close .= "\n";
      else
        $msg_close = "connection finished\n";

      echo '= '.$this->sonic_stream_sr_prelog().' ';
      echo $msg_close;
    }

    if ($this->sock_mux_out)
      socket_close($this->sock_mux_out);

    if ($this->sock_out)
      socket_close($this->sock_out);

    if ($this->sock);
      socket_close($this->sock);
  }

  /**
   * stream main
   *
   * @see SONIC_stream::sonic_stream_main()
   * @param resource $sock_in third party support (library mode) override, resource socket bound and listening for connections
  */
  function sonic_stream_main($sock_in = null)
  {
    //non threaded protection
    if (function_exists('_sonic_stream_sr_sockout'));
    else
    {
//
    function _sonic_stream_sr_sockout($_this)
    {
      $config = $_this->config;
      //DST
      $stream_config = $config['stream_config'];

      if ($host = $stream_config['dst_host']);
      else
        die('! '.$_this->sonic_stream_sr_prelog().' main() failed: reason: destination host not defined in config'."\n");

      if ($port = $stream_config['dst_port']) $port += 0;
      else
        die('! '.$_this->sonic_stream_sr_prelog().' main() failed: reason: destination port not defined in config'."\n");

      $src_ip = $stream_config['src_ip'];

      //SETUP CONNECTION OUT AND BIND SRC IP
      //helper suff. (setup for inbound socket type stream)
      if ($rs_domain = $stream_config['relay_socket_domain'])
      { 
        if ($_def = @constant($rs_domain))
          $rs_domain = $_def;
      }
      else if ($rs_domain = $config['socket_domain'])
      { 
        if ($_def = @constant($rs_domain))
          $rs_domain = $_def;
      }
      else
        $rs_domain = SONIC_SOCK_DOMAIN; //eg. AF_INET
      if ($rs_type = $stream_config['relay_socket_type'])
      { 
        if ($_def = @constant($rs_type))
          $rs_type = $_def;
      }
      else if ($rs_type = $config['socket_type'])
      { 
        if ($_def = @constant($rs_type))
          $rs_type = $_def;
      }
      else
        $rs_type = SONIC_SOCK_TYPE; //eg. SOCK_STREAM
      if ($rs_protocol = $stream_config['relay_socket_protocol'])
      { 
        if ($_def = @constant($rs_protocol))
          $rs_protocol = $_def;
      }
      else if ($rs_protocol = $config['socket_protocol'])
      {
        if ($_def = @constant($rs_protocol))
          $rs_protocol = $_def;
      }
      else
        $rs_protocol = SONIC_SOCK_PROTOCOL; //eg. SOL_TCP
      if ($rs_options = $stream_config['relay_socket_options'])
      { 
        if ($_def = @constant($rs_options))
          $rs_options = $_def;
      }
      else if ($rs_options = $config['socket_options'])
      {
        if ($_def = @constant($rs_options))
          $rs_options = $_def;
      }
      else
        $rs_options = SONIC_SOCK_OPTIONS;  //eg. reuseaddr=1,

      unset($_def);

      //string tcp -> SOL_TCP map check
      if (gettype($rs_protocol) == 'string')
      {
        if (($_rsp = getprotobyname($rs_protocol)) == -1)
          $rs_protocol = SONIC_SOCK_PROTOCOL;
        else
        {
          $rs_protocol = $_rsp;
          unset($_rsp);
        }
      }

      if (($sock_out = @socket_create($rs_domain, $rs_type, $rs_protocol)) == false)
        die('! '.$_this->sonic_stream_sr_prelog().' socket_create() failed: reason: '.socket_strerror(socket_last_error())."\n");

      sonic_socket_set_options($sock_out, SOL_SOCKET, $rs_options);

      if ($src_ip)
        if (@socket_bind($sock_out, $src_ip) == false)
          die('! '.$_this->sonic_stream_sr_prelog().' socket_bind() failed: reason: '.socket_strerror(socket_last_error($sock_out))."\n");

      if ($rs_type == SOCK_STREAM)
      {
        if (@socket_connect($sock_out, $host, $port) == false)
          die('! '.$_this->sonic_stream_sr_prelog().' socket_connect() failed: reason: '.socket_strerror(socket_last_error($sock_out))."\n");
      }

      //MUX
      if ($host_mux = $stream_config['mux_host']);
      else
        unset($host_mux);

      if ($port_mux = $stream_config['mux_port']) $mux_port += 0;
      else
        unset($port_mux);

      if ($port_mux == NULL);
      else
      {
        if ($src_ip_mux = $stream_config['mux_src_ip']);
        else if ($src_ip_mux = $src_ip);
        else
          unset($src_ip_mux);

        //SETUP CONNECTION OUT AND BIND SRC IP
        //helper suff. (setup for inbound socket type stream)
        if ($rsm_domain = $stream_config['mux_socket_domain'])
        { 
          if ($_def = @constant($rsm_domain))
            $rsm_domain = $_def;
        }
        else if ($rsm_domain = $rs_domain)
        { 
          if ($_def = @constant($rsm_domain))
            $rsm_domain = $_def;
        }
        else if ($rsm_domain = $config['socket_domain'])
        { 
          if ($_def = @constant($rsm_domain))
            $rsm_domain = $_def;
        }
        else
          $rsm_domain = SONIC_SOCK_DOMAIN; //eg. AF_INET
        if ($rsm_type = $stream_config['mux_socket_type'])
        { 
          if ($_def = @constant($rsm_type))
            $rsm_type = $_def;
        }
        else if ($rsm_type = $rs_type)
        { 
          if ($_def = @constant($rsm_type))
            $rsm_type = $_def;
        }
        else if ($rsm_type = $config['socket_type'])
        { 
          if ($_def = @constant($rsm_type))
            $rsm_type = $_def;
        }
        else
          $rsm_type = SONIC_SOCK_TYPE; //eg. SOCK_STREAM
        if ($rsm_protocol = $stream_config['mux_socket_protocol'])
        { 
          if ($_def = @constant($rsm_protocol))
            $rsm_protocol = $_def;
        }
        else if ($rsm_protocol = $rs_protocol)
        { 
          if ($_def = @constant($rsm_protocol))
            $rsm_protocol = $_def;
        }
        else if ($rsm_protocol = $config['socket_protocol'])
        { 
          if ($_def = @constant($rsm_protocol))
            $rsm_protocol = $_def;
        }
        else
          $rsm_protocol = SONIC_SOCK_PROTOCOL; //eg. SOL_TCP
        if ($rsm_options = $stream_config['mux_socket_options'])
        { 
          if ($_def = @constant($rsm_options))
            $rsm_options = $_def;
        }
        else if ($rsm_options = $rs_options)
        { 
          if ($_def = @constant($rsm_options))
            $rsm_options = $_def;
        }
        else if ($rsm_options = $config['socket_options'])
        { 
          if ($_def = @constant($rsm_options))
            $rsm_options = $_def;
        }
        else
          $rsm_options = SONIC_SOCK_OPTIONS; //eg. reuseaddr=1,

        unset($_def);

        //string tcp -> SOL_TCP map check
        if (gettype($rsm_protocol) == 'string')
        {
          if (($_rsmp = getprotobyname($rsm_protocol)) == -1)
            $rsm_protocol = SONIC_SOCK_PROTOCOL;
          else
          {
            $rsm_protocol = $_rsmp;
            unset($_rsmp);
          }
        }

        if (($sock_out_mux = @socket_create($rsm_domain, $rsm_type, $rsm_protocol)) == false)
          die('! '.$_this->sonic_stream_sr_prelog().' socket_create() failed: reason: '.socket_strerror(socket_last_error())."\n");

        sonic_socket_set_options($sock_out_mux, SOL_SOCKET, $rsm_options);

        if ($src_ip_mux)
          if (@socket_bind($sock_out_mux, $src_ip_mux) == false)
            die('! '.$_this->sonic_stream_sr_prelog().' socket_bind() failed: reason: '.socket_strerror(socket_last_error($sock_out_mux))."\n");

        if ($rsm_type == SOCK_STREAM)
          if (@socket_connect($sock_out_mux, $host_mux, $port_mux) == false)
            die('! '.$_this->sonic_stream_sr_prelog().' socket_connect() failed: reason: '.socket_strerror(socket_last_error($sock_out_mux))."\n");
      }

      $sr_obj = array('socket' 		=> $sock_out,
		 'socket_domain' 	=> $rs_domain,
		 'socket_type' 		=> $rs_type,
		 'socket_protocol' 	=> $rs_protocol,
		 'socket_options' 	=> $rs_options,
		 'host' 		=> $host,
		 'port' 		=> $port
		);

      if ($sock_out_mux)
      {
        $sr_obj['mux_socket'] 		= $sock_out_mux;
        $sr_obj['mux_socket_domain'] 	= $rsm_domain;
        $sr_obj['mux_socket_type'] 	= $rsm_type;
        $sr_obj['mux_socket_protocol'] 	= $rsm_protocol;
        $sr_obj['mux_socket_options'] 	= $rsm_options;
        $sr_obj['mux_host'] 		= $host_mux;
        $sr_obj['mux_port'] 		= $port_mux;
      }

      return $sr_obj;
    }
//
    } //nonthreaded protection

    $this->sonic_stream_sr_start();

    $sock_in = &$this->sock;

    //quit after accept
//    if ((($this->pid == 0 && $this->pid_child == 0) OR ($this->pid == $this->pid_child)) && $this->rotate_shutdown)
    if ($this->rotate_shutdown)
    { 
      $this->sonic_stream_sr_stop();
      return;
    }

    $stream_config = $this->config['stream_config'];

    static $sock_in_write_init = false;

    do
    { 
      $this->sonic_stream_buffer();

      $ss_read = array($sock_in);
      $ss_write = NULL;
      $ss_except = NULL;
      $ss_changed_sockets = socket_select($ss_read, $ss_write, $ss_except, 0, 1);
      if ($ss_changed_sockets > 0 OR $this->config['socket_type'] != SOCK_STREAM)
      {
  //READ IN SEND TO OUT
        if (false == ($buf_in = socket_read($sock_in, SONIC_RELAY_BYTE_SIZE, PHP_BINARY_READ)))
          break;
        else if ($this->verbose > 998 OR $this->verbose == 22)
          echo $buf_in;

        if ($sock_out);
        else //sock_in input came first
        {
          $_sr_socko = _sonic_stream_sr_sockout($this);
          $sock_out = &$_sr_socko['socket'];
          $this->sock_out = &$sock_out; 
          $sock_domain = $_sr_socko['socket_domain'];
          $sock_type = $_sr_socko['socket_type'];
          $sock_protocol = $_sr_socko['socket_protocol'];
          $sock_options = $_sr_socko['socket_options'];
          $host = $_sr_socko['host'];
          $port = $_sr_socko['port'];

          if ($_sr_socko['mux_socket'])
          {
            $sock_out_mux = &$_sr_socko['mux_socket'];
            $this->sock_mux_out = &$sock_out_mux;
            $sock_domain_mux = $_sr_socko['mux_socket_domain'];
            $sock_type_mux = $_sr_socko['mux_socket_type'];
            $sock_protocol_mux = $_sr_socko['mux_socket_protocol'];
            $sock_options_mux = $_sr_socko['mux_socket_options'];
            $host_mux = $_sr_socko['mux_host'];
            $port_mux = $_sr_socko['mux_port'];
          }
          else
            unset($sock_out_mux); //noidz trigger safety
        }

        if ($sock_type == SOCK_STREAM)
          $o_bytes = @socket_write($sock_out, $buf_in, strlen($buf_in));
        else
          $o_bytes = @socket_sendto($sock_out, $buf_in, strlen($buf_in), 0, $host, $port);

        if ($o_bytes < 0 OR $o_bytes == false)
          echo '! '.$this->sonic_stream_sr_prelog().' main() echo failed: reason: '.socket_strerror(socket_last_error($sock_out))."\n";
      }

      //non TCP sock_out would have init already above after first input on sock_in
      if ($sock_out);
      else
      {
        $_sr_socko = _sonic_stream_sr_sockout($this);
        $sock_out = &$_sr_socko['socket'];
        $this->sock_out = &$sock_out;
        $sock_domain = $_sr_socko['socket_domain'];
        $sock_type = $_sr_socko['socket_type'];
        $sock_protocol = $_sr_socko['socket_protocol'];
        $sock_options = $_sr_socko['socket_options'];
        $host = $_sr_socko['host'];
        $port = $_sr_socko['port'];

        if ($_sr_socko['mux_socket'])
        {
          $sock_out_mux = &$_sr_socko['mux_socket'];
          $this->sock_mux_out = &$sock_out_mux;
          $sock_domain_mux = $_sr_socko['mux_socket_domain'];
          $sock_type_mux = $_sr_socko['mux_socket_type'];
          $sock_protocol_mux = $_sr_socko['mux_socket_protocol'];
          $sock_options_mux = $_sr_socko['mux_socket_options'];
          $host_mux = $_sr_socko['mux_host'];
          $port_mux = $_sr_socko['mux_port'];
        }
        else
          unset($sock_out_mux); //noidz trigger safety
      }

      //if two way street on both sides
      if ($sock_in_write_init);
      else if ($sock_type == SOCK_STREAM) //sock_out can posstbly return data
      {
        if ($sock_out_mux OR $this->config['socket_type'] == SOCK_STREAM)
          $sock_in_write_init = true;
      }
      if ($sock_in_write_init)
      {
        $ss_read = array($sock_out);
        $ss_write = NULL;
        $ss_except = NULL;
        $ss_changed_sockets = socket_select($ss_read, $ss_write, $ss_except, 0, 1);
        if ($ss_changed_sockets > 0)
        {
    //READ OUT SEND TO IN
          if (false == ($buf_out = socket_read($sock_out, SONIC_RELAY_BYTE_SIZE, PHP_BINARY_READ)))
            break;
          else if ($this->verbose > 998 OR $this->verbose == 51)
            echo $buf_out;

          if ($sock_out_mux)
          {
            if ($sock_type_mux == SOCK_STREAM)
              $i_bytes = @socket_write($sock_out_mux, $buf_out, strlen($buf_out));
            else
              $i_bytes = @socket_sendto($sock_out_mux, $buf_out, strlen($buf_out), 0, $host_mux, $port_mux);

            if ($i_bytes < 0 OR $i_bytes == false)
              echo '! '.$this->sonic_stream_sr_prelog().' main() echo failed: reason: '.socket_strerror(socket_last_error($sock_out_mux))."\n";
          }
          else
          {
            $i_bytes = @socket_write($sock_in, $buf_out, strlen($buf_out));

            if ($i_bytes < 0 OR $i_bytes == false)
              echo '! '.$this->sonic_stream_sr_prelog().' main() echo failed: reason: '.socket_strerror(socket_last_error($sock_in))."\n";
          }
        }
      }

//      if ((($this->pid == 0 && $this->pid_child == 0) OR ($this->pid == $this->pid_child)) && $this->rotate_shutdown)
      if ($this->rotate_shutdown)
        break;

    } while (true);

    $this->sonic_stream_sr_stop();
  }

  /**
   * log output unifier
   *
   * returns a uniform prefix for log/output entries
   *
   * @see SONIC_stream::sonic_stream_prelog()
   * @return string string usually with pid/process id and timestamp
  */
  function sonic_stream_sr_prelog()
  {
    return $this->sonic_stream_prelog();
  }
}

?>
